from .entry import *

from .exit import *

__all__ = ['TaiChiLearn',
            'ALL_ENTRY', 'ALL_EXIT', 'QUANTIFY_2_ENTRY_MAP', 'QUANTIFY_2_EXIT_MAP',
           'EntryModel', 'Empty', 'ImageConvEncoder',
           'CategoryEncoder', 'MultiCategoryEncoder',
           'TransformerEncoder',
           'RegressionTop', 'CategoryTop', 'MultiCategoryTop']




__all__ = ["CategoryTop", "MultiCategoryTop"]

from .metric import accuracy, bi_accuracy
from .loss import BCEWithLogitsLossCasted
from .basic import ExitModel, nn


class CategoryTop(ExitModel):
    prefer = "CrossEntropyLoss"
    input_dim = 2

    def __init__(self, in_features, out_features):
        super().__init__()
        self.top = nn.Linear(
            in_features=in_features, out_features=out_features)
        self.activation = nn.Softmax(dim=-1)
        self.crit = nn.CrossEntropyLoss()
        self.metric_funcs.update({"acc": accuracy})

    def forward(self, x):
        return self.top(x)

    @classmethod
    def from_quantify(cls, quantify, entry_part):
        out_features = len(quantify.category)
        in_features = entry_part.out_features
        return cls(
            in_features=in_features,
            out_features=out_features,
        )


class MultiCategoryTop(ExitModel):
    prefer = "BCEWithLogitsLossCasted"
    input_dim = 2

    def __init__(self, in_features, out_features):
        super().__init__()
        self.top = nn.Linear(
            in_features=in_features, out_features=out_features)
        self.activation = nn.Sigmoid()
        self.crit = BCEWithLogitsLossCasted()
        self.metric_funcs.update({"acc": bi_accuracy})

    def forward(self, x):
        return self.top(x)

    @classmethod
    def from_quantify(cls, quantify, entry_part):
        out_features = len(quantify.category)
        in_features = entry_part.out_features
        return cls(
            in_features=in_features,
            out_features=out_features,
        )



__all__ = ["QUANTIFY","Quantify", "QuantifyImage", "QuantifyText",
           "QuantifyCategory", "QuantifyMultiCategory", "QuantifyNum"]

from .basic import (
    Quantify, QuantifyImage, QuantifyText, 
    QuantifyCategory, QuantifyMultiCategory, QuantifyNum
    )

QUANTIFY = dict(
    Quantify=Quantify,
    QuantifyNum=QuantifyNum,
    QuantifyImage=QuantifyImage,
    QuantifyCategory=QuantifyCategory,
    QuantifyMultiCategory=QuantifyMultiCategory,
    QuantifyText=QuantifyText,
)


__all__ = ["ExitModel", "RegressionTop"]

from torch import nn
import numpy as np


class ExitModel(nn.Module):
    metric_funcs = dict()

    def loss_step(self, x, y):
        y_ = self(x)
        loss = self.crit(y_, y)
        metrics = dict()
        if hasattr(self, "activation"):
            y_ = self.activation(y_)
        for k, func in self.metric_funcs.items():
            metrics.update({k: func(y_, y)})
        return dict(loss=loss, y_=y_, **metrics)

    def eval_forward(self, x) -> np.ndarray:
        """
        Forward pass for 
        """
        y_ = self(x)
        if hasattr(self, "activation"):
            y_ = self.activation(y_)
        return y_

    @classmethod
    def from_quantify(cls, ):
        raise ImportError(
            f"Please define class function 'from_quantify' for {cls.__name__}"
        )


class RegressionTop(ExitModel):
    prefer = "MSELoss"
    input_dim = 2

    def __init__(self, in_features, out_features):
        super().__init__()
        self.top = nn.Linear(
            in_features=in_features, out_features=out_features)
        self.crit = nn.MSELoss()

    def forward(self, x):
        return self.top(x)

    @classmethod
    def from_quantify(cls, quantify, entry_part):
        out_features = 1
        in_features = entry_part.out_features
        return cls(
            in_features=in_features,
            out_features=out_features,
        )



__all__ = ["Enrich", "EnrichCleanTyping"]

from pathlib import Path
import json
from tai_chi_engine.utils import clean_name
from tai_chi_engine.stateful import Stateful
from typing import Union
from tai_chi_tuna.front.typer import STR, LIST


class Enrich(Stateful):
    phase_state = 'enrich'
    """
    Enrich Base Class
    Some default attributes
    - is_enrich = True
    - typing = None # output typing
    - multi_cols = False # use multi-column as input
    - prefer = None
    - lazy = False  # shall we execute enrichment only through the iteration
    - src = None # source column
    """
    is_enrich = True
    typing = None  # output typing
    multi_cols = False  # use multi-column as input
    prefer = None
    lazy = False  # shall we execute enrichment only through the iteration
    src = None  # source column

    # which properties to save to disk
    stateful_conf = dict()

    def __init__(self): pass

    def __call__(self, row):
        return row

    def rowing(self, row):
        if self.multi_cols:
            return self(row)
        else:
            return self(row[self.src])

class EnrichCleanTyping(Enrich):
    """
    Typing a column, and fill in the missing value
    The src and the dst can be same column for this case
    """
    stateful_conf = dict(to_type='str', missing='str')
    typing = Union[str, int, float, bool] # output typing
    def __init__(
        self,
        to_type: LIST(options=['String','Float','Integer','Bool'], default="Float")='Float',
        missing: STR(default="0") = "0",
        ):
        self.to_type = to_type
        self.missing = missing

        self.the_type = dict({'String':str,'Float':float,'Integer':int,'Bool':bool})[to_type]
        self.mst = self.the_type(missing)

    def __repr__(self,):
        return f"[CleanTyping:{self.to_type}, with missing value to: {self.missing}]"

    def __call__(self, x):
        if x is None:
            # return the default missing value
            return self.mst
        # return the typed x
        return self.the_type(x)



__all__ = ["Stateful", "STATE_TYPE"]

from pathlib import Path
from tai_chi_engine.utils import clean_name
import json


class StateType:
    def __init__(self, type_):
        self.type_ = type_

    def __repr__(self):
        return f"{self.type_}"

    def load(self, save_dir, value):
        return self.type_(value)

    def save(self, save_dir, value):
        """
        Return the value that can be saved as JSON string
        """
        return value


class LoadCategory(StateType):
    def __init__(self):
        super().__init__("Category")

    def __repr__(self):
        return f"{self.filename}"

    def load(self, save_dir, value):
        from forgebox.category import Category
        return Category.load(save_dir/value)

    def save(self, save_dir, value) -> str:
        value.save(save_dir/"category.json")
        return "category.json"


class LoadPretrainedTokenizer(StateType):
    def __init__(self):
        super().__init__("Transformers.Tokenizer")

    def __repr__(self):
        return f"{self.filename}"

    def load(self, save_dir, value):
        from transformers import AutoTokenizer
        tokenizer = AutoTokenizer.from_pretrained(
            str(save_dir/value), use_fast=True)
        return tokenizer

    def save(self, save_dir, value) -> str:
        pretrained = "tokenizer"
        from transformers import AutoTokenizer
        value.save_pretrained(str(save_dir/pretrained))
        return pretrained


STATE_TYPE = dict({
    "str": StateType(str),
    "int": StateType(int),
    "float": StateType(float),
    "bool": StateType(bool),
    "list": StateType(list),
    "category": LoadCategory(),
    "tokenizer": LoadPretrainedTokenizer(),
})


class Stateful:
    phase_state = "tai-chi-engine"
    stateful_conf = dict()

    @classmethod
    def phase_state_dir(cls, project: Path) -> Path:
        """
        phase state directory under project directory
        """
        project = Path(project)
        phase_state = project/cls.phase_state
        if not phase_state.exists():
            phase_state.mkdir(exist_ok=True, parents=True)
        return phase_state

    @classmethod
    def save_dir(cls, project: Path, name: str) -> Path:
        """
        The directory to save the column
        """
        save_dir = cls.phase_state_dir(project)/clean_name(name)
        if not save_dir.exists():
            save_dir.mkdir(exist_ok=True, parents=True)
        return save_dir

    def save(self, project: Path, name: str):
        """
        default save function
        """
        save_dir = self.save_dir(project, name)
        to_save_data = dict()
        # iter through property name and property type
        for property_name, state_type_key in self.stateful_conf.items():
            running_value = getattr(self, property_name)
            state_type = STATE_TYPE[state_type_key]
            save_value = state_type.save(save_dir, running_value)
            to_save_data[property_name] = save_value
        with open(save_dir/"config.json", "w") as f:
            json.dump(to_save_data, f)

    @classmethod
    def load_conf(cls, project: Path, name: str):
        """
        Load object from configuration file
        """
        save_dir = cls.save_dir(project, name)
        config_path = save_dir/'config.json'
        if config_path.exists():
            with open(save_dir/"config.json", 'r') as f:
                config = json.loads(f.read())
        else:
            config = dict()
        return config

    @classmethod
    def load(cls, project: Path, name: str):
        """
        Load from configuration file
        """
        save_dir = cls.save_dir(project, name)
        config = cls.load_conf(project, name)

        # property data for __init__ function
        init_dict = dict()
        # property data to set attribute after __init__
        setattr_dict = dict()
        # translate the config to the stateful object
        for property_name, state_type_key in cls.stateful_conf.items():
            state_type = STATE_TYPE[state_type_key]
            running_value = state_type.load(
                save_dir, config[property_name])

            if property_name in cls.__init__.__annotations__:
                init_dict[property_name] = running_value
            else:
                setattr_dict[property_name] = running_value
        obj = cls(**init_dict)
        for property_name, running_value in setattr_dict.items():
            setattr(obj, property_name, running_value)
        return obj



__all__ = ['TransformerEncoder']

from tai_chi_tuna.front.typer import BOOL, STR
from .basic import EntryModel

class TransformerEncoder(EntryModel):
    """
    A model part to encode sequnce data in to vectors
    """

    def __init__(self, model, encoder_mode: BOOL(default=True) = True,):
        super().__init__()
        self.model = model
        self.encoder_mode = encoder_mode

    def forward(self, kwargs):
        outputs = self.model(**kwargs)
        if self.encoder_mode:
            # output vector
            if "pooler_output" in outputs:
                return outputs.pooler_output
            else:
                return (
                    outputs.last_hidden_state*kwargs['attention_mask'][:,:,None]
                ).mean(1)
        return outputs

    @classmethod
    def from_quantify(
        cls,
        quantify,
        name: STR(default="bert-base-uncased") = 'bert-base-uncased',
        encoder_mode: BOOL(default=True) = True,
    ):
        from transformers import AutoModel
        model = AutoModel.from_pretrained(name)
        obj = cls(model)
        obj.name = name
        obj.encoder_mode = encoder_mode
        if encoder_mode:
            obj.out_features= model.config.hidden_size
        return obj


__all__ = ["Quantify", "QuantifyImage", "QuantifyText",
           "QuantifyCategory", "QuantifyMultiCategory", "QuantifyNum"]

from tai_chi_tuna.front.html import Flash
from tai_chi_tuna.front.typer import STR, INT, BOOL, LIST

from tai_chi_tuna.flow.to_quantify import (
    BATCH_SIZE, SEQUENCE_SIZE, IMAGE_SIZE)
from tai_chi_engine.stateful import Stateful

from forgebox.category import Category
from typing import List
import numpy as np
import pandas as pd
import torch
import json
from pathlib import Path


class Quantify(Stateful):
    phase_state = "quantify"
    is_quantify = True
    stateful_conf = dict()
    """
    # From all things to number
    The AI model does not understand anything, say, picture, text
    Unless you transform it to integer and float tensors

    Quantify and its subclass controls the
        numericalization / collation of the data pipeline
    The base class of quantify does: NOTHING
    """

    def __init__(self,):
        pass

    def __call__(self, list_of_items):
        return list(list_of_items)

    def adapt(self, column):
        """
        A function to let the data processing
        adapt to the data column
        """
        pass

    def __hash__(self,):
        if hasattr(self, "name"):
            return self.name
        else:
            return self.__class__.__name__

    def backward(self, x):
        """
        convert tensor back to the original data
        """
        return x


class QuantifyImage(Quantify):
    stateful_conf = dict(
        mean_='list',
        std_='list',
    )
    """
    Transform PIL.Image to tensor
    """

    def __init__(
        self,
        mean_: LIST(["imagenet", "0.5 x 3"]) = "imagenet",
        std_: LIST(["imagenet", "0.5 x 3"]) = "imagenet",
    ):
        if type(mean_) == str:
            if mean_ == "imagenet":
                mean_ = [0.485, 0.456, 0.406]
            elif mean_ == "0.5 x 3":
                mean_ = [.5, .5, .5]
            else:
                raise ValueError(
                    f"Mean configuration: {mean_} not valid")

        if type(std_) == str:
            if std_ == "imagenet":
                std_ = [0.229, 0.224, 0.225]
            elif std_ == "0.5 x 3":
                std_ = [.5, .5, .5]
            else:
                raise ValueError(
                    f"Standard Variation configuration: {std_} not valid")

        from torchvision import transforms as tfm
        self.transform = tfm.Compose([
            tfm.ToTensor(),
            tfm.Normalize(mean=mean_, std=std_),
        ])
        self.mean_ = mean_
        self.std_ = std_

        self.shape = (BATCH_SIZE, 3, IMAGE_SIZE, IMAGE_SIZE)

    def __repr__(self):
        return f"Quantify Image to tensors:{self.transform}"

    def __call__(self, list_of_image):
        return torch.stack(list(
            self.transform(img) for img in list_of_image))


class QuantifyText(Quantify):
    stateful_conf = dict(
        pretrained='str',
        max_length='int',
        padding='str',
        return_token_type_ids='bool',
        return_attention_mask='bool',
        return_offsets_mapping='bool',
        tokenizer='tokenizer',
    )
    """
    Tokenize, numercialize, pad the texts to tensor
    """

    def __init__(
        self,
        pretrained: STR(default="bert-base-cased") = "bert-base-cased",
        max_length: INT(default=512, min_=12, max_=1024, step=4) = 512,
        padding: LIST(options=[
            "do_not_pad",
            "max_length",
            "longest"], default="max_length") = "max_length",
        return_token_type_ids: BOOL(
            name="Token Type IDs", default=True) = True,
        return_attention_mask: BOOL(
            name="Attention Mask", default=True) = True,
        return_offsets_mapping: BOOL(
            name="Offset Mapping", default=False) = False,
    ):
        self.pretrained = pretrained
        self.max_length = max_length
        self.padding = padding
        self.return_token_type_ids = return_token_type_ids
        self.return_attention_mask = return_attention_mask
        self.return_offsets_mapping = return_offsets_mapping
        self.truncation = True
        self.return_tensors = 'pt'
        self.shape = (BATCH_SIZE, SEQUENCE_SIZE)

    def adapt(self, column):
        """
        Initialize tokenizer
        """
        from transformers import AutoTokenizer
        Flash.info("Loading transformer tokenizer, takes time", key="Alert!")
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.pretrained, use_fast=True)

    def __call__(self, list_of_text: List[str]):
        list_of_text = list(list_of_text)
        return self.tokenizer(
            list_of_text,
            padding=self.padding,
            max_length=self.max_length,
            truncation=self.truncation,
            return_token_type_ids=self.return_token_type_ids,
            return_attention_mask=self.return_attention_mask,
            return_tensors=self.return_tensors,
            return_offsets_mapping=self.return_offsets_mapping,
        )


class QuantifyCategory(Quantify):
    stateful_conf = dict(
        category='category',
        min_frequency='int',
    )
    """
    Transform single categorical data to index numbers in pytorch tensors
    """

    def __init__(
        self,
        min_frequency: INT(min_=1, max_=20, default=1) = 1,
    ):
        self.min_frequency = min_frequency

    def adapt(self, column):
        # category statistics
        value_counts = pd.DataFrame(column.value_counts())

        # if minimun freq is 1
        # very category occured should be accounted for
        # hence no missing token padding is required
        if self.min_frequency < 2:
            self.category = Category(
                arr=np.array(value_counts.index),
                pad_mst=False)

        # we need missing token
        # for category's frequency < self.min_frequency
        else:
            categories = np.array(
                list(value_counts.index[
                    value_counts.values.reshape(-1) > self.min_frequency]))
            self.category = Category(arr=categories, pad_mst=True)

    def __repr__(self):
        return f"Quantify Category:{self.category}"

    def __call__(self, list_of_strings):
        return torch.LongTensor(self.category.c2i[np.array(list_of_strings)])

    def backward(self, array):
        """
        Reconstruct a row of data
        """
        array = array.cpu().numpy()
        df = pd.DataFrame(dict(category=self.category.i2c, score = array))
        df = df.sort_values(by='score', ascending=False).reset_index(drop=True)
        return df


class QuantifyMultiCategory(Quantify):
    stateful_conf = dict(
        category='category',
        min_frequency='int',
        separator='str',
    )
    """
    Transform multi-categorical data to index numbers in pytorch tensors
    """

    def __init__(
        self,
        min_frequency: INT(min_=1, max_=20, default=1) = 1,
        separator: LIST(options=["[None]", ",", ";",
                        "[Space]", "[By Char]"], default=",") = ",",
    ):
        self.min_frequency = min_frequency
        friendly_mapping = {
            "[None]": None, # the input will be a list already
            "[Space]": " ",
            "[By Char]": "",
        }
        self.separator = separator
        if separator in friendly_mapping:
            actual = friendly_mapping.get(separator)
            self.actual_separator = actual
        else:
            self.actual_separator = separator

    @staticmethod
    def stripping(x):
        return x.strip()

    def break_cell(self, value):
        if value is None:
            return []
        break_list = list(i for i in map(
            self.stripping, str(value).split(self.actual_separator)) if len(i) > 0)
        return break_list

    def adapt(self, column):
        if self.actual_separator is None:
            sample_col = column
        else:
            sample_col = column.apply(self.break_cell)
        value_counts = pd.DataFrame(sample_col.explode().value_counts())

        # if minimun freq is 1
        # very category occured should be accounted for
        # hence no missing token padding is required
        if self.min_frequency < 2:
            self.category = Category(
                arr=np.array(value_counts.index),
                pad_mst=False)

        # we need missing token
        # for category's frequency < self.min_frequency
        else:
            categories = np.array(
                list(value_counts.index[
                    value_counts.values.reshape(-1) > self.min_frequency]))
            self.category = Category(arr=categories, pad_mst=True)

    def __call__(
        self, list_of_strings: List[str]
    ) -> torch.LongTensor:
        """
        Return a batch of n-hot array tensor
        """
        if self.actual_separator is None:
            col: List[str] = list_of_strings
        else:
            col: List[List[str]] = list(map(self.break_cell, list_of_strings))
        arrays: List[np.array] = []
        for item in col:
            array: np.array = np.zeros(len(self.category))
            if len(item) > 0:
                one_idx: np.array = self.category.c2i[item]
                array[one_idx] = 1
            arrays.append(array)
        return torch.LongTensor(np.stack(arrays))

    def backward(self, array):
        """
        Reconstruct a row of prediction to categorical information
        """
        array = array.cpu().numpy()
        df = pd.DataFrame(dict(category=self.category.i2c, score = array))
        df = df.sort_values(by='score', ascending=False).reset_index(drop=True)
        return df


class QuantifyNum(Quantify):
    stateful_conf = dict(
        mean_='float',
        std_='float',
    )
    """
    Quantify contineous data, like float numbers
    The only process is normalization on the entire population
    """
    shape = (BATCH_SIZE, 1)

    def adapt(self, column):
        self.mean_ = column.mean()
        self.std_ = column.std()

    def __call__(self, list_of_num):
        return (torch.FloatTensor(list_of_num)[:, None]-self.mean_)/self.std_

    def backward(self, x):
        """
        Reconstruct the value to the scale of the original data
        """
        return (x*self.std_+self.mean_).cpu().numpy()


__all__ = ["TaiChiEngine",]
from tai_chi_tuna.flow.trunk import TaiChiLearn
from .enrich import ENRICHMENTS
from .quantify import QUANTIFY
from .model import (ALL_EXIT, QUANTIFY_2_EXIT_MAP,
                    ALL_ENTRY, QUANTIFY_2_ENTRY_MAP)


class TaiChiEngine(TaiChiLearn):
    """
    TaiChiEngine
    One and the only thing the end user has to initiate and operate.
    """
    
    enrichments_map = ENRICHMENTS
    quantify_map = QUANTIFY
    quantify_2_entry_map = QUANTIFY_2_ENTRY_MAP
    quantify_2_exit_map = QUANTIFY_2_EXIT_MAP
    all_entry = ALL_ENTRY
    all_exit = ALL_EXIT

